@@ -25,11 +25,11 @@ from .. import mocpy
 from .boundaries import Boundaries
 from .plot import fill, border
 
-__author__ = "Thomas Boch, Matthieu Baumann"
+__author__ = "Thomas Boch, Matthieu Baumann, F.-X. Pineau"
 __copyright__ = "CDS, Centre de Données astronomiques de Strasbourg"
 
 __license__ = "BSD 3-Clause License"
-__email__ = "thomas.boch@astro.unistra.fr, matthieu.baumann@astro.unistra.fr"
+__email__ = "thomas.boch@astro.unistra.fr, baumannmatthieu0@gmail.com, francois-xavier.pineau@astro.unistra.fr"
 
 class MOC(AbstractMOC):
     """
@@ -66,9 +66,83 @@ class MOC(AbstractMOC):
     5. Serialize `~mocpy.moc.MOC` objects to `astropy.io.fits.HDUList` or JSON dictionary and save it to a file.
     """
 
-    def __init__(self, interval_set=None):
+    # I introduced, but do not like, the double `make_consistent` (MOC + IntervalSet)
+    # but `coverage_merge_time_intervals` is no more genric
+    # and I can't remove `make_consistent` from `IntervalSet` without changing tests
+    def __init__(self, interval_set=None, make_consistent=True, min_depth=None):
+        """
+        Moc constructor.
+
+        The merging step of the overlapping intervals is done here.
+
+        Parameters
+        ----------
+        intervals : `~numpy.ndarray`
+            a N x 2 numpy array representing the set of intervals.
+        make_consistent : bool, optional
+            True by default. Remove the overlapping intervals that makes
+            a valid MOC (i.e. can be plot, serialized, manipulated).
+        """
         super(MOC, self).__init__(interval_set)
 
+        if make_consistent:
+            if min_depth is None:
+                min_depth = -1
+
+            min_depth = np.int8(min_depth)
+            self._merge_intervals(min_depth)
+
+    def _merge_intervals(self, min_depth):
+        if not self.empty():
+            self._interval_set._intervals = mocpy.coverage_merge_hpx_intervals(self._interval_set._intervals, min_depth)
+
+    @property
+    def max_order(self):
+        """
+        Depth of the smallest HEALPix cells found in the MOC instance.
+        """
+        depth = mocpy.hpx_coverage_depth(self._interval_set._intervals)
+        depth = np.uint8(depth)
+        return depth
+
+    def refine_to_order(self, min_depth):
+        intervals = mocpy.coverage_merge_hpx_intervals(self._interval_set._intervals, min_depth)
+        interval_set = IntervalSet(intervals, make_consistent=False)
+        return MOC(interval_set, make_consistent=False)
+
+    def complement(self):
+        """
+        Returns the complement of the MOC instance.
+
+        Returns
+        -------
+        result : `~mocpy.moc.MOC`
+            The resulting MOC.
+        """
+        intervals = mocpy.hpx_coverage_complement(self._interval_set._intervals)
+        interval_set = IntervalSet(intervals, make_consistent=False)
+        return MOC(interval_set, make_consistent=False)
+
+    def degrade_to_order(self, new_order):
+        """
+        Degrades the MOC instance to a new, less precise, MOC.
+
+        The maximum depth (i.e. the depth of the smallest HEALPix cells that can be found in the MOC) of the
+        degraded MOC is set to ``new_order``.
+
+        Parameters
+        ----------
+        new_order : int
+            Maximum depth of the output degraded MOC.
+
+        Returns
+        -------
+        moc : `~mocpy.moc.MOC`
+            The degraded MOC.
+        """
+        intervals = mocpy.hpx_coverage_degrade(self._interval_set._intervals, new_order)
+        return MOC(IntervalSet(intervals, make_consistent=False), make_consistent=False)
+
     def contains(self, ra, dec, keep_inside=True):
         """
         Returns a boolean mask array of the positions lying inside (or outside) the MOC instance.
@@ -115,30 +189,11 @@ class MOC(AbstractMOC):
         moc : `~mocpy.moc.MOC`
             self extended by one degree of neighbours.
         """
-        max_depth = self.max_order
-
-        # Get the pixels array of the MOC at the its max order.
-        ipix = mocpy.flatten_pixels(self._interval_set._intervals, max_depth)
-        # Get the HEALPix array containing the neighbors of ``ipix``.
-        # This array "extends" ``ipix`` by one degree of neighbors.
-        ipix_extended = cdshealpix.neighbours(ipix, max_depth)
-        ipix_extended = ipix_extended[ipix_extended >= 0]
-        ipix_extended = ipix_extended.astype(np.uint64)
-
-        # Compute the difference between ``extend_ipix`` and ``ipix`` to get only the neighboring pixels
-        # located at the border of the MOC.
-        ipix_neighbors = np.setdiff1d(ipix_extended, ipix)
-
-        depth_neighbors = np.full(shape=ipix_neighbors.shape, fill_value=max_depth, dtype=np.int8)
-        #intervals_neighbors = mocpy.from_healpix_cells(ipix_neighbors, depth_neighbors)
-        moc_neighbors = MOC.from_healpix_cells(ipix_neighbors, depth_neighbors)
-        
-        # This array of HEALPix neighbors are added to the MOC to get an ``extended`` MOC
-        #self._interval_set._intervals = mocpy.coverage_union(self._interval_set._intervals, moc_neighbors._interval_set._intervals)
-        final = self.union(moc_neighbors)
-        self._interval_set = final._interval_set
+        intervals = mocpy.hpx_coverage_expand(self.max_order, self._interval_set._intervals)
+        self._interval_set = IntervalSet(intervals, make_consistent=False)
         return self
 
+
     def remove_neighbours(self):
         """
         Removes from the MOC instance the HEALPix cells located at its border.
@@ -150,29 +205,11 @@ class MOC(AbstractMOC):
         moc : `~mocpy.moc.MOC`
             self minus its HEALPix cells located at its border.
         """
-        max_depth = self.max_order
-        # Get the HEALPix cells of the MOC at its max depth
-        ipix = mocpy.flatten_pixels(self._interval_set._intervals, max_depth)
-        # Get the HEALPix array containing the neighbors of ``ipix``.
-        # This array "extends" ``ipix`` by one degree of neighbors.
-        ipix_extended = cdshealpix.neighbours(ipix, max_depth)
-        ipix_extended = ipix_extended[ipix_extended >= 0]
-        ipix_extended = ipix_extended.astype(np.uint64)
-        # Get only the max depth HEALPix cells lying at the border of the MOC
-        ipix_neighbors = np.setxor1d(ipix_extended, ipix)
-
-        # Remove these pixels from ``ipix``
-        ipix_around_border = cdshealpix.neighbours(ipix_neighbors, max_depth)
-        ipix_around_border = ipix_around_border[ipix_around_border >= 0]
-        ipix_around_border = ipix_around_border.astype(np.uint64)
-
-        final_ipix = np.setdiff1d(ipix, ipix_around_border)
-        final_depth = np.full(shape=final_ipix.shape, fill_value=max_depth, dtype=np.int8)
-
-        # Build the reduced MOC, i.e. MOC without its pixels which were located at its border.
-        final = MOC.from_healpix_cells(final_ipix, final_depth)
-        self._interval_set = final._interval_set
+        intervals = mocpy.hpx_coverage_contract(self.max_order, self._interval_set._intervals);
+        self._interval_set = IntervalSet(intervals, make_consistent=False)
         return self
+        
+
 
     def fill(self, ax, wcs, **kw_mpl_pathpatch):
         """
@@ -514,10 +551,13 @@ class MOC(AbstractMOC):
             The resulting MOC
         """
         intervals = mocpy.from_lonlat(max_norder, lon.to_value(u.rad).astype(np.float64), lat.to_value(u.rad).astype(np.float64))
-        return cls(IntervalSet(intervals, make_consistent=False))
+        return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
 
     @classmethod
-    def from_valued_healpix_cells(cls, uniq, values, max_depth=None, cumul_from=0.0, cumul_to=1.0):
+    def from_valued_healpix_cells(cls, 
+        uniq, values, max_depth=None, 
+        cumul_from=0.0, cumul_to=1.0, 
+        asc=False, strict=True, no_split=True, reverse_decent=False):
         """
         Creates a MOC from a list of uniq associated with values.
 
@@ -527,6 +567,11 @@ class MOC(AbstractMOC):
         Cells being on the fence are recursively splitted and added
         until the depth of the cells is equal to ``max_norder``.
 
+        For compatibility with Aladin, use ``no_split=False`` and ``reverse_decent=True``
+        
+        Remark: using ``no_split=False``, the way the cells overlapping with the low and high thresholds are split
+        is somewhat arbitrary.
+
         Parameters
         ----------
         uniq : `numpy.ndarray`
@@ -540,6 +585,14 @@ class MOC(AbstractMOC):
             Cumulative value from which cells will be added to the MOC
         cumul_to : float
             Cumulative value to which cells will be added to the MOC
+        asc: boolean
+            the cumulative value is computed from lower to highest densities instead of from highest to lowest
+        strict: boolean
+            (sub-)cells overlapping the `cumul_from` or `cumul_to` values are not added
+        no_split: boolean
+            cells overlapping the `cumul_from` or `cumul_to` values are not recursively split
+        reverse_decent: boolean
+            perform the recursive decent from the highest cell number to the lowest (to be compatible with Aladin)
 
         Returns
         -------
@@ -560,9 +613,13 @@ class MOC(AbstractMOC):
             uniq.astype(np.uint64),
             values.astype(np.float64),
             np.float64(cumul_from),
-            np.float64(cumul_to)
+            np.float64(cumul_to),
+            asc, 
+            strict, 
+            no_split, 
+            reverse_decent
         )
-        moc = cls(IntervalSet(intervals, make_consistent=False))
+        moc = cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
 
         # Degrade the MOC to the depth requested by the user
         if max_depth is not None:
@@ -755,8 +812,8 @@ class MOC(AbstractMOC):
         if fully_covered is not None and fully_covered.shape != ipix.shape:
             raise IndexError("fully covered and depth arrays must have the same shape")
 
-        intervals = mocpy.from_healpix_cells(ipix.astype(np.uint64), depth.astype(np.int8))
-        return cls(IntervalSet(intervals, make_consistent=False))
+        intervals = mocpy.from_healpix_cells(ipix.astype(np.uint64), depth.astype(np.uint8))
+        return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
 
     @staticmethod
     def order_to_spatial_resolution(order):
@@ -886,7 +943,7 @@ class MOC(AbstractMOC):
         warnings.simplefilter('default')
         warnings.warn('This method is deprecated and is no longer tested.'
                       'Please refer to this documentation page for plotting MOCs using'
-                      'matplotlib: https://mocpy.readthedocs.io/en/latest/examples/examples.html#loading-and-plotting-the-moc-of-sdss', DeprecationWarning)
+                      'matplotlib: https://cds-astro.github.io/mocpy/xamples/examples.html#loading-and-plotting-the-moc-of-sdss', DeprecationWarning)
 
         frame = ICRS() if frame is None else frame
 
@@ -941,3 +998,102 @@ class MOC(AbstractMOC):
         plt.grid(True, linestyle='--', linewidth=1, color='#555555')
 
         plt.show()
+
+    def save(self, path, format='fits'):
+        """
+        Writes the Spatial MOC to a file.
+
+        Format can be 'fits', 'ascii', or 'json', though the json format is not officially supported by the IVOA.
+
+        Parameters
+        ----------
+        path : str
+            The path to the file to save the MOC in.
+        format : str, optional
+            The format in which the MOC is saved.
+            Possible formats are "fits", "ascii" or "json".
+            By default, ``format`` is set to "fits".
+        """
+        if format == 'fits':
+            mocpy.spatial_moc_to_fits_file(self.max_order, self._interval_set._intervals, path)
+        elif format == 'ascii':
+            mocpy.spatial_moc_to_ascii_file(self.max_order, self._interval_set._intervals, path)
+        elif format == 'json':
+            mocpy.spatial_moc_to_json_file(self.max_order, self._interval_set._intervals, path)
+        else:
+            formats = ('fits', 'ascii', 'json')
+            raise ValueError('format should be one of %s' % (str(formats)))
+
+    @classmethod
+    def load(cls, path, format='fits'):
+        """
+        Load the Spatial MOC from a file.
+
+        Format can be 'fits', 'ascii', or 'json', though the json format is not officially supported by the IVOA.
+
+        Parameters
+        ----------
+        path : str
+            The path to the file to load the MOC from.
+        format : str, optional
+            The format from which the MOC is loaded.
+            Possible formats are "fits", "ascii" or "json".
+            By default, ``format`` is set to "fits".
+        """
+        if format == 'fits':
+            intervals = mocpy.spatial_moc_from_fits_file(path)
+            return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
+        elif format == 'ascii':
+            intervals = mocpy.spatial_moc_from_ascii_file(path)
+            return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
+        elif format == 'json':
+            intervals = mocpy.spatial_moc_from_json_file(path)
+            return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
+        else:
+            formats = ('fits', 'ascii', 'json')
+            raise ValueError('format should be one of %s' % (str(formats)))
+
+    def to_string(self, format='ascii'):
+        """
+        Writes the Spatial MOC into a string.
+
+        Format can be 'ascii' or 'json', though the json format is not officially supported by the IVOA.
+
+        Parameters
+        ----------
+        format : str, optional
+            The format in which the MOC will be serialized before being saved.
+            Possible formats are "ascii" or "json".
+            By default, ``format`` is set to "ascii".
+        """
+        if format == 'ascii':
+            return mocpy.spatial_moc_to_ascii_str(self.max_order, self._interval_set._intervals)
+        elif format == 'json':
+            return mocpy.spatial_moc_to_json_str(self.max_order, self._interval_set._intervals)
+        else:
+            formats = ('ascii', 'json')
+            raise ValueError('format should be one of %s' % (str(formats)))
+
+    @classmethod
+    def from_string(cls, value, format='ascii'):
+        """
+        Deserialize the Spatial MOC from the given string.
+
+        Format can be 'ascii' or 'json', though the json format is not officially supported by the IVOA.
+
+        Parameters
+        ----------
+        format : str, optional
+            The format in which the MOC will be serialized before being saved.
+            Possible formats are "ascii" or "json".
+            By default, ``format`` is set to "ascii".
+        """
+        if format == 'ascii':
+            intervals = mocpy.spatial_moc_from_ascii_str(value)
+            return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
+        elif format == 'json':
+            intervals = mocpy.spatial_moc_from_json_str(value)
+            return cls(IntervalSet(intervals, make_consistent=False), make_consistent=False)
+        else:
+            formats = ('ascii', 'json')
+            raise ValueError('format should be one of %s' % (str(formats)))
\ No newline at end of file
